from enum import Enum
from typing import Any, Optional

import numpy as np
import pandas as pd

from exodusutils.enums import DataType
from exodusutils.exceptions.exceptions import ExodusForbidden, ExodusMethodNotAllowed
from exodusutils.internal.process_unit import ProcessUnit
from exodusutils.schemas import Column


class Method(str, Enum):
    """
    The currently supported imputation methods.
    """

    mode = "mode"
    mean = "mean"
    average = "average"
    zero = "zero"
    min = "min"
    max = "max"
    median = "median"


class Imputation(ProcessUnit):
    """
    Imputes the given column in the dataframe with a specified method.

    Available methods are:
    - "mode": Only supports non-empty columns
    - "mean": Only supports numeric columns
    - "average": Alias of "mean" method
    - "zero": Imputes numeric columns with 0, empty string "" for other types of columns
    - "min"
    - "max"
    - "median": Only supports numeric columns

    """

    def __init__(self, column: Column, method: Method) -> None:
        self.column = column
        self.method = method
        self.target: Optional[Any] = None

    def fitted(self) -> bool:
        return self.target is not None

    def fit(self, df: pd.DataFrame) -> None:
        values = df[df[self.column.name].notnull()][self.column.name]
        if self.method == Method.mode:
            mode = values.mode()[0]
            if mode:
                target = mode
            else:
                raise ExodusMethodNotAllowed(
                    f"Cannot calculate mode of empty column = {self.column.name}"
                )
        elif self.method == Method.mean or self.method == Method.average:
            if self.column.data_type == DataType.double:
                target = values.mean()
            else:
                raise ExodusMethodNotAllowed(
                    f"Imputing categorical column = {self.column.name} with method = {self.method.value} is not supported"
                )
        elif self.method == Method.zero:
            if self.column.data_type == DataType.double:
                target = 0
            else:
                target = ""
        elif self.method == Method.min:
            target = values.min()
        elif self.method == Method.max:
            target = values.max()
        elif self.method == Method.median:
            if self.column.data_type == DataType.double:
                target = values.median()
            else:
                raise ExodusMethodNotAllowed(
                    f"Imputing categorical column = {self.column.name} with method = {self.method.value} is not supported"
                )
        else:
            raise ExodusMethodNotAllowed(
                f"Unsupported imputation method = {self.method.value}"
            )
        if self.column.data_type == DataType.double and np.isnan(target):
            raise ExodusForbidden(
                f"Cannot impute column = {self.column.name} with np.nan"
            )
        self.target = target

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        if not self.fitted():
            raise ExodusMethodNotAllowed("Has to fit before transform")
        return df.fillna(value={self.column.name: self.target})



from enum import Enum
from typing import List, Optional

import pandas as pd
from pydantic import BaseModel
from pydantic.class_validators import validator
from pydantic.fields import Field

from exodusutils import internal
from exodusutils.enums import DataType
from exodusutils.exceptions.exceptions import ExodusBadRequest
from exodusutils.schemas import Column


class TrainReqBodyBase(BaseModel):
    """
    The base schema for training requests.

    Attributes
    ----------
    training_data : bytes
        The training data as a sequence of bytes.
    feature_types : List[Column]
        The features present in training data.
    target : Column
        The target column. Must be present in `feature_types`.
    folds : int
        Number of folds for this experiment.
    """

    training_data: bytes = Field(description="The training data as a sequence of bytes")
    feature_types: List[Column] = Field(
        default=[], description="The features present in training data"
    )
    target: Column = Field(
        description="The target column. Must be present in feature_types"
    )
    folds: int = Field(
        default=5, ge=2, le=10, description="Number of folds for this experiment"
    )

    @validator("target")
    def check_column(cls, v, values):
        if v not in values.get("feature_types", []):
            raise ExodusBadRequest(
                f"target {v} not found in feature_types {values.get('feature_types', [])}"
            )
        return v

    @validator("feature_types")
    def validate_header(cls, v, values):
        data = values["training_data"]
        header = data.decode("utf-8").split("\n")[0].split(",")
        internal.validate_columns(header, v)
        return v

    def get_feature_names(self) -> List[str]:
        """
        Returns the names of the features.
        """
        return [f.name for f in self.feature_types]

    def get_training_df(self) -> pd.DataFrame:
        """
        Returns the Pandas DataFrame from the bytes in training_data field.
        """
        return internal.get_df(self.training_data, self.feature_types)

    def get_features(self, data_type: DataType) -> List[Column]:
        """
        Returns the features with the specified data type.

        Parameters
        ----------
        data_type : DataType
            The specified data type.

        Returns
        -------
        List[Column]
            The features.

        """
        return [f for f in self.feature_types if f.data_type == data_type]

    def get_feature_names_by_data_type(self, data_type: DataType) -> List[str]:
        """
        Helper function for extracting names for features with a specified type.

        Parameters
        ----------
        data_type : DataType
            The specified data type.

        Returns
        -------
        List[str]
            Names of the features.
        """
        return [f.name for f in self.get_features(data_type)]


class TrainIIDReqBody(TrainReqBodyBase):
    """
    The schema for IID training request body.

    Has the following fields:
    - `feature_types`: a list of `Column`s, which is the name of a feature and the feature's data type. Contains the target.
    - `target`: the target `Column`.
    - `folds`: number of cross validation folds we want to train.
    - `random_seed`: random seed. Optional.

    Get dataframes by calling the following methods:
    - `get_training_df`: returns the training data as a Pandas dataframe.
    - `get_holdout_df`: returns the holdout data as a Pandas dataframe if the field is not a `None`. Otherwise returns a `None`.
    - `get_validation_df`: returns the validation data as a Pandas dataframe if the field is not a `None`. Otherwise returns a `None`.
    """

    random_seed: Optional[int] = Field(
        default=None, description="The random seed for this experiment. Optional."
    )
    validation_data: Optional[bytes] = Field(
        default=None, description="The validation data as a sequence of bytes. Optional"
    )
    holdout_data: Optional[bytes] = Field(
        default=None, description="The holdout data as a sequence of bytes. Optional"
    )
    fold_assignment_column_name: Optional[str] = Field(
        default=None,
        description="The name of the fold assignment column. If not provided, Exodus will cut cross validation folds in a modulo fashion. If this field is defined, it is required to be a valid column in the input dataframe. This column is not included in `feature_types`, and will be discarded during training.",
    )

    @validator("fold_assignment_column_name")
    def validate_fold_column(cls, v, values):
        if v is not None:
            data = values["training_data"]
            header = data.decode("utf-8").split("\n")[0].split(",")
            if v not in header:
                raise ExodusBadRequest(
                    f"fold assignment column = {v} not found in header"
                )
        return v

    class Config:
        schema_extra = {
            "examples": [
                {
                    "training_data": "foo,bar\n0,abc\n1,def\n2,ghi\n3,jkl\n4,mnop\n5,qrs\n",
                    "feature_types": [
                        {"name": "foo", "data_type": "double"},
                        {"name": "bar", "data_type": "string"},
                    ],
                    "target": [{"name": "bar", "data_type": "string"}],
                    "folds": 3,
                    "validation_percentage": 0.2,
                    "holdout_data": "foo,bar\n99,zzz\n",
                }
            ]
        }

    def get_holdout_df(self) -> Optional[pd.DataFrame]:
        """
        Returns the Pandas DataFrame from the bytes in holdout_data field, or `None` if none specified.
        """
        if not self.holdout_data:
            return None
        else:
            return internal.get_df(self.holdout_data, self.feature_types)

    def get_validation_df(self) -> Optional[pd.DataFrame]:
        """
        Returns the Pandas DataFrame from the bytes in holdout_data field, or `None` if none specified.
        """
        if not self.validation_data:
            return None
        else:
            return internal.get_df(self.validation_data, self.feature_types)


class PredictReqBody(BaseModel):
    """
    The schema for prediction request body.
    """

    model_id: str = Field(description="The specified model id")
    data: str = Field(description="The prediction data as JSON")
    threshold: Optional[float] = Field(
        default=None, description="The threshold for classification predictions"
    )
    keep_columns: List[str] = Field(
        default=[], description="The columns to keep in the prediction response"
    )

    def get_prediction_df(self, feature_types: List[Column]) -> pd.DataFrame:
        """
        Returns the Pandas DataFrame from the bytes in data field.

        If the input data contains any column that is not in `feature_types`, that column will be \
dropped from the resulting dataframe.
        """
        valid_columns = [f.name for f in feature_types]
        df = pd.DataFrame(pd.read_json(self.data, orient="records"))
        valid_features = [f for f in feature_types if f.name in df.columns.tolist()]

        # We only want to keep the columns that are specified in `feature_types`
        return df.drop(
            [c for c in df.columns.tolist() if c not in valid_columns], axis=1
        ).pipe(internal.cast_df_types, valid_features)


class MigrateAction(str, Enum):
    """
    Defines the migration action. Could only be either one of `up` and `down`.
    """

    up = "up"
    down = "down"


class MigrateReqBody(BaseModel):
    """
    The schema for migrate request body.
    """

    action: MigrateAction = Field(description="The migration action")

    class Config:
        use_enum_values = True



from typing import List, Optional

import numpy as np
import pandas as pd

from exodusutils import internal
from exodusutils.exceptions.exceptions import ExodusMethodNotAllowed
from exodusutils.internal.process_unit import ProcessUnit


class OneHotEncoding(ProcessUnit):
    """
    Does one-hot encoding. If a dummy column that we found during `fit` is not there during `transform`, \
we create the column in the to-be-transformed dataframe, and fill it with `0`.
    """

    def __init__(self, target_column_name: str) -> None:
        self.target_column_name = target_column_name
        self.dummy_columns: Optional[List[str]] = None

    def fitted(self) -> bool:
        return self.dummy_columns is not None

    def fit(self, df: pd.DataFrame) -> None:
        categorical_columns = [
            c for c in internal.get_columns(df, object) if c != self.target_column_name
        ]
        dummy = pd.get_dummies(df, columns=categorical_columns, dtype=np.int64)
        self.dummy_columns = [col for col in dummy.columns.to_list() if col not in df]

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        if not self.fitted() or self.dummy_columns is None:
            raise ExodusMethodNotAllowed("Has to fit before transform")
        categorical_columns = [
            c for c in internal.get_columns(df, object) if c != self.target_column_name
        ]
        dummy = pd.get_dummies(df, columns=categorical_columns, dtype=np.int64)
        for c in self.dummy_columns:
            if c not in dummy.columns:
                dummy[c] = 0
        return dummy



from typing import Any, Dict, List, Optional

from pydantic import BaseModel
from pydantic.fields import Field

from exodusutils.schemas import Attribute
from exodusutils.schemas.scores import CVScores, Scores


class Attributes(BaseModel):
    cv_scores: CVScores
    variable_importance: List[Attribute] = []
    validation_scores: Optional[CVScores] = None
    holdout_scores: Optional[Scores] = None


class TrainRespBody(BaseModel):
    """
    The schema for train response body.
    """

    name: str = Field(description="The name of the model algorithm")
    model_id: str = Field(description="The specified model id")
    hyperparameters: Dict[str, Any] = Field(description="The hyperparameters")
    attributes: Attributes = Field(
        description="The attributes, including CV scores, validation scores, etc."
    )


class PredictRespBody(BaseModel):
    """
    The schema for prediction response body.
    """

    prediction: List[Dict[str, Any]] = Field(description="The prediction response")


class InfoRespBody(BaseModel):
    """
    The schema for info response body.
    """

    name: str = Field(description="Name of the model algorithm")
    description: str = Field(description="Description of the model algorithm")



__version__ = "0.1.0"

from typing import List

import numpy as np
import pandas as pd
from Levenshtein import distance

from exodusutils.enums import TimeUnit


def closest_string(words: List[str], ref: str) -> str:
    """
    Returns the word in `words` that has the minimum Levenshtein distance from `ref`.
    """
    return sorted(words, key=lambda w: distance(w, ref))[0]


def remove_invalid_targets(df: pd.DataFrame, target: str) -> pd.DataFrame:
    """
    Returns a new dataframe that does not have `NaN` in target column. Does not modify the original dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        df
    target : str
        target

    Returns
    -------
    pd.DataFrame

    """
    return df.loc[df[target].notnull(), :]


def format_datetime_column_with_unit(
    df: pd.DataFrame, datetime_column: str, time_unit: TimeUnit
) -> pd.DataFrame:
    """Turns date column into `str` based on given `time_unit`. Will modify the original `df`."""
    df[datetime_column] = df[datetime_column].apply(
        lambda t: time_unit.format_datetime(t)
    )
    return df


def are_valid_values(values: List[float]) -> bool:
    """
    Whether there exists a non-`np.nan` value in `values`. Used to test if a fold is useless - \
if all actual values are `np.nan`, then the fold is not usable.

    Parameters
    ----------
    values : List[float]
       The values

    Returns
    -------
    bool

    """
    return not all([np.isnan(x) for x in values])



from pydantic import BaseModel

from exodusutils.enums import DataType


class Attribute(BaseModel):
    """A kv pair representing an attribute. Used for feature importances."""

    name: str
    value: float


class Column(BaseModel):
    """Represents a column in the dataframe."""

    name: str
    data_type: DataType

    class Config:
        use_enum_values = True



from exodusutils.preprocessing.column_cardinality_limiter import (
    ColumnCardinalityLimiter,
)
from exodusutils.preprocessing.constant_columns_remover import ConstantColumnsRemover
from exodusutils.preprocessing.imputation import Imputation, Method



from typing import List

import numpy as np
import pandas as pd

from exodusutils import internal
from exodusutils.enums import TimeUnit
from exodusutils.internal.process_unit import ProcessUnit


class TimeComponentEncoding(ProcessUnit):
    """
    Does time component encoding. The `fit` method does nothing. For `transform`, the encoding extract the \
following:
    - year, to a column named `\"{COL_NAME}_y\"`
    - quarter, to a column named `\"{COL_NAME}_q\"`
    - month, to a column named `\"{COL_NAME}_m\"`
    - week, to a column named `\"{COL_NAME}_w\"`
    - weekday, to a column named `\"{COL_NAME}_d\"`
    - hour, to a column named `\"{COL_NAME}_h\"`
    - minute, to a column named `\"{COL_NAME}_m\"`
    - second, to a column named `\"{COL_NAME}_s\"`
    And then it would calculate the difference in seconds from the values in the column to the Unix Epoch (1970-01-01), \
and substitute the column with the difference.

    """

    def __init__(self, time_unit: TimeUnit = TimeUnit.hour) -> None:
        self.time_unit = TimeUnit(time_unit)
        self.components: List[str] = []

    def fit(self, df: pd.DataFrame) -> None:
        pass

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        self.components = []
        for column in internal.get_columns(df, np.datetime64):
            for prefixed_column_name, component in zip(
                internal.postfix_datetime_cols_with_time_components(column),
                internal.get_time_components(df, self.time_unit, column),
            ):
                if component is not None:
                    df[prefixed_column_name] = component
                    self.components.append(prefixed_column_name)
            df[column] = (df[column] - pd.Timestamp("1970-01-01")) // pd.Timedelta("1s")
        return df



from typing import Optional

from fastapi import HTTPException, status


class ExodusBadRequest(HTTPException):
    def __init__(self, detail: Optional[str] = None) -> None:
        super().__init__(status_code=status.HTTP_400_BAD_REQUEST, detail=detail)


class ExodusNotFound(HTTPException):
    def __init__(self, detail: Optional[str] = None) -> None:
        super().__init__(status_code=status.HTTP_404_NOT_FOUND, detail=detail)


class ExodusForbidden(HTTPException):
    def __init__(self, detail: Optional[str] = None) -> None:
        super().__init__(status_code=status.HTTP_403_FORBIDDEN, detail=detail)


class ExodusMethodNotAllowed(HTTPException):
    def __init__(self, detail: Optional[str] = None) -> None:
        super().__init__(status_code=status.HTTP_405_METHOD_NOT_ALLOWED, detail=detail)


class ExodusError(HTTPException):
    def __init__(self, detail: Optional[str] = None) -> None:
        super().__init__(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=detail
        )



from enum import Enum
from typing import Dict, Optional

import numpy as np
import pandas as pd
from sklearn.preprocessing import LabelEncoder

from exodusutils import internal
from exodusutils.exceptions.exceptions import ExodusMethodNotAllowed
from exodusutils.internal.process_unit import ProcessUnit


class HandleInvalidLabel(str, Enum):
    new_label = "new_label"
    invalid_label = "invalid_label"


class LabelEncoding(ProcessUnit):
    """
    Does label encoding. After fitting, objects of this class would contain a valid `encoders`, which includes \
the `sklearn.preprocessing.LabelEncoder` for each encoded columns.

    To handle unseen labels during `transform`, 2 methods are currently supported:
        - Use a label specifying invalid label (`-1`) to represent them. This is the default behavior.
        - Use an additional label (`len(encoder.classes_)`) to represent them.

    You can change the behavior while initializing this class.
    """

    def __init__(
        self,
        target_column_name: str,
        handle_invalid_label: HandleInvalidLabel = HandleInvalidLabel.invalid_label,
    ) -> None:
        self.target_column_name = target_column_name
        self.handle_invalid_label = handle_invalid_label
        self.encoders: Optional[Dict[str, LabelEncoder]] = None

    def fitted(self) -> bool:
        return self.encoders is not None

    def fit(self, df: pd.DataFrame) -> None:
        encoders = dict()
        for name in internal.get_columns(df, object):
            if name != self.target_column_name:
                encoders[name] = LabelEncoder().fit(
                    df.loc[df[name].notnull(), name].astype(str)
                )
        self.encoders = encoders

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        if not self.fitted() or self.encoders is None:
            raise ExodusMethodNotAllowed("Has to fit before transform")
        for name, encoder in self.encoders.items():
            df.loc[
                ~df[name].isin(encoder.classes_), name
            ] = np.nan  # unseen values in training data
            df.loc[df[name].notnull(), name] = encoder.transform(
                df.loc[df[name].notnull()][name]
            )
            df.loc[df[name].isna(), name] = (
                -1
                if self.handle_invalid_label.value == HandleInvalidLabel.invalid_label
                else len(encoder.classes_)
            )  # if it's never seen before, just fill it with `-1`, which means the label is invalid
            df[name] = pd.to_numeric(pd.Series(df[name]), errors="coerce")
        return df



from typing import Dict, List, Optional

import numpy as np
import pandas as pd

from exodusutils.enums import DataType
from exodusutils.exceptions.exceptions import ExodusMethodNotAllowed
from exodusutils.internal.process_unit import ProcessUnit
from exodusutils.schemas import Column


class ColumnCardinalityLimiter(ProcessUnit):
    """
    Limits the number of distinct labels within one categorical column. Will not limit the target column.

    The default max. cardinality is 30. Any other categorical labels will be turned into the same `\"InvalidLabel\"` label.
    """

    def __init__(
        self,
        feature_types: List[Column],
        target_column_name: str,
        max_cardinality: int = 30,
        invalid_label: str = "InvalidLabel",
    ) -> None:
        self.feature_types = feature_types
        self.target_column_name = target_column_name
        self.max_cardinality = max_cardinality
        self.invalid_label = invalid_label
        self.cardinalities: Optional[Dict[str, List[str]]] = None

    def fitted(self) -> bool:
        return self.cardinalities is not None

    def fit(self, df: pd.DataFrame) -> None:
        self.cardinalities = {}
        for column_name in [
            c.name
            for c in self.feature_types
            if c.data_type == DataType.string
            and df[c.name].nunique() > self.max_cardinality
            and c.name != self.target_column_name
        ]:
            self.cardinalities[column_name] = (
                df[column_name].value_counts().index[: self.max_cardinality].to_list()
            )

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        if not self.fitted() or self.cardinalities is None:
            raise ExodusMethodNotAllowed("Has to fit before transform")
        for column_name, valid_labels in self.cardinalities.items():
            df.loc[
                (~df[column_name].isin(valid_labels)) & (df[column_name].notnull()),
                [column_name],
            ] = self.invalid_label
        return df



from datetime import datetime
from enum import Enum

from pandas.api.types import is_datetime64_any_dtype, is_numeric_dtype
from pydantic.main import BaseModel


class TimeUnit(str, Enum):
    """
    Represents a time unit.
    """

    hour = "hour"
    day = "day"
    month = "month"
    year = "year"

    def to_seasonality(self) -> int:
        """
        How many time units there is in a recurring pattern.

        For example, the seasonality of time unit `Month` would be `12`, since it's 12 months in a year.
        """
        if self.value == TimeUnit.hour:
            return 24
        elif self.value == TimeUnit.day:
            return 7
        elif self.value == TimeUnit.month:
            return 12
        else:
            return 1

    def to_resample_rule(self) -> str:
        """
        Turns the time unit to a string literal, which will then be used by pandas to resample the dataset.
        """
        if self.value == TimeUnit.year:
            return "YS"
        elif self.value == TimeUnit.month:
            return "MS"
        elif self.value == TimeUnit.day:
            return "D"
        else:
            return "H"

    def to_timedelta_unit(self) -> str:
        if self.value == TimeUnit.year:
            return "Y"
        elif self.value == TimeUnit.month:
            return "M"
        elif self.value == TimeUnit.day:
            return "D"
        else:
            return "h"

    def format_datetime(self, date: datetime) -> str:
        """Format a `datetime` based on this time unit.

        Parameters
        ----------
        date : datetime
            The `datetime` object to format.

        Returns
        -------
        The formatted `datetime` object as a `str`.
        """
        if self.value == TimeUnit.year:
            return date.strftime("%Y-01-01")
        elif self.value == TimeUnit.month:
            return date.strftime("%Y-%m-01")
        elif self.value == TimeUnit.day:
            return date.strftime("%Y-%m-%d")
        else:
            return date.strftime("%Y-%m-%dT%H:00:00")

    def to_prediction_format(self) -> str:
        """The corresponding format for parsing the dates in the prediction input data set."""
        if self.value == TimeUnit.hour:
            return "%Y-%m-%d %H:%M:%S"
        else:
            return "%Y-%m-%d"

    def to_last_date_format(self) -> str:
        """The corresponding format for parsing the last date in the training set."""
        if self.value == TimeUnit.hour:
            return "%Y-%m-%dT%H:%M:%S"
        else:
            return "%Y-%m-%d"


class ID(BaseModel):
    value: str


class DataType(str, Enum):
    string = "string"
    double = "double"
    timestamp = "timestamp"
    id = "id"

    def to_type(self):
        if self.value == DataType.string:
            return str
        elif self.value == DataType.double:
            return float
        elif self.value == DataType.timestamp:
            return datetime
        elif self.value == DataType.id:
            return ID
        else:
            raise RuntimeError("unprocessable data type")

    @classmethod
    def from_pandas(cls, arr_or_dtype):
        """
        Converts from Pandas types to our custom `DataType`.

        Can take either an array-ish input, or the pandas column dtype.
        """
        if is_numeric_dtype(arr_or_dtype):
            return cls.double
        elif is_datetime64_any_dtype(arr_or_dtype):
            return cls.timestamp
        else:
            return cls.string


class AlgoType(str, Enum):
    TS = "TS"
    IID = "IID"



DATETIME_COMPONENTS_PREFIX = (
    "y",  # year
    "q",  # quarter
    "m",  # month
    "w",  # week
    "d",  # day of week
    "h",  # hour
    "mm",  # minute
    "s",  # second
)

PREDICTION_COLNAME = "prediction"



import abc

from pymongo.collection import Collection


class BaseMigration(abc.ABC):
    """
    The base class for migration scripts.

    User will need to implement the `up` and `down` methods.
    """

    def __init__(self, timestamp: int, name: str) -> None:
        """
        Initialize this migration script.

        Parameters
        ----------
        timestamp : int
            The timestamp as an `int`.
        name : str
            The name of this migration script.
        """
        self.timestamp = timestamp
        self.name = name
        self.mongo_name = f"{self.timestamp}_{self.name}"

    @abc.abstractmethod
    def up(self, collection: Collection) -> None:
        """
        The up step of this migration script.

        Parameters
        ----------
        collection : Collection
            The MongoDB collection this script will operate on.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def down(self, collection: Collection) -> None:
        """
        The down step of this migration script. Should either do the inverse
        of the `up` step, or do nothing if the `up` step is irreversible.

        Parameters
        ----------
        collection : Collection
            The MongoDB collection this script will operate on.
        """
        raise NotImplementedError



import logging
import math
from typing import Callable, DefaultDict, Dict, List, Union

import numpy as np
import numpy.typing as npt
from pydantic import BaseModel, validator
from pydantic.types import NonNegativeFloat
from sklearn import preprocessing
from sklearn.metrics import accuracy_score, auc, log_loss, mean_squared_error, roc_curve

from exodusutils.internal import sanitize_pred_and_actual


class RegressionScores(BaseModel):
    """
    The regression scores set.

    Metrics include:
        - mean squared error (`mse`)
        - root mean square error (`rmse`)
        - root mean square logarithmic error (`rmsle`)
        - mean absolute error (`mae`)
        - r squared (`r2`)
        - deviance (`deviance`)
        - mean absolute percentage error (`mape`)
        - weighted mean absolute percentage error (`wmape`)
    """

    mse: float
    rmse: float
    rmsle: float
    mae: float
    r2: float
    deviance: float
    mape: float
    wmape: float

    @validator("mse", "rmse", "rmsle", "mae", "mape", "wmape")
    def check_nonnegative_or_nan(cls, v):
        if not math.isnan(v) and v < 0:
            raise ValueError(
                f"Invalid score: required nan or non-negative float, got {v}"
            )
        return v

    @classmethod
    def get_scores(cls, pred: npt.NDArray[np.float64], actual: npt.NDArray[np.float64]):
        """
        Calculate the regrssion scores according to a set of predicted values (`pred`) and actual values (`actual`).

        Parameters
        ----------
        pred : np.ndarray
            Prediction results based on test set feature columns
        actual : np.ndarray
            The test set target values from the original dataframe

        Returns
        -------
        `RegressionScores`
        """
        pred, actual = sanitize_pred_and_actual(pred, actual)

        variance = np.var(actual)

        def calc_rmsle(y: np.ndarray, yhat: np.ndarray) -> float:
            try:
                log_diffs = list()
                for i in range(len(yhat)):
                    cur_pred = math.log(y[i] + 1)
                    cur_act = math.log(yhat[i] + 1)
                    log_diffs.append(pow(cur_act - cur_pred, 2))
                msle = sum(log_diffs) / len(log_diffs)
                res = math.sqrt(msle)
            except Exception as e:
                logging.error("Unable to calculate rmsle: " + str(e))
                res = np.NaN
            return res

        def calc_r2(yhat: np.ndarray, y: np.ndarray) -> float:
            diffs = np.abs(np.subtract(yhat, y))
            diffs_not_nan = diffs[~np.isnan(diffs)]

            square_diffs = [np.square(x) for x in diffs_not_nan]
            mse = sum(square_diffs) / len(square_diffs)

            return 1 - mse / variance

        def calc_metric_score(f: Callable[[np.ndarray, np.ndarray], float]):
            if actual.shape != pred.shape:
                raise RuntimeError(
                    f"actual shape ({actual.shape}) != predicted shape ({pred.shape})"
                )
            shape = actual.shape
            if len(shape) == 1:  # single target
                res = f(pred, actual)
            elif len(shape) == 2:  # multiple targets
                ls = list()
                for i in range(shape[0]):
                    ls.append(f(pred[i], actual[i]))
                res = sum(ls) / len(ls)
            else:  # can't happen
                raise RuntimeError(f"invalid shape: {shape}")
            return res

        mse = np.round(mean_squared_error(pred, actual), 5)
        rmse = np.round(np.sqrt(mse), 5)
        absolute_difference = np.abs(np.subtract(actual, pred))
        mae = np.round(np.mean(absolute_difference), 5)
        mape = (
            float("nan")
            if 0 in actual
            else np.round(100 * np.mean(absolute_difference / actual), 5)
        )
        wmape = (
            float("nan")
            if np.sum(actual) == 0.0
            else np.round(100 * np.sum(absolute_difference) / np.sum(actual), 5)
        )
        r2 = np.round(calc_metric_score(calc_r2), 5)
        deviance = mse
        rmsle = np.round(calc_metric_score(calc_rmsle), 5)
        return cls(
            mse=mse,
            rmse=rmse,
            rmsle=rmsle,
            mae=mae,
            mape=mape,
            wmape=wmape,
            r2=r2,
            deviance=deviance,
        )


class ClassificationScores(BaseModel):
    """
    If there are more than 2 classes, it is a MultinomialScores. Otherwise it is a BinomialScores.

    For BinomialScores, the metrics include:
        - logloss (`logloss`)
        - mean per class error (`mean_per_class_error`)
        - misclassification (`misclassification`)
        - area under curve (`auc`)
        - lift top group (`lift_top_group`)

    For MultinomialScores, the metrics include:
        - logloss (`logloss`)
        - mean per class error (`mean_per_class_error`)
        - misclassification (`misclassification`)

    """

    logloss: float
    mean_per_class_error: NonNegativeFloat
    misclassification: NonNegativeFloat

    @staticmethod
    def get_scores(
        pred: np.ndarray, pred_proba: np.ndarray, actual: np.ndarray, labels: np.ndarray
    ):
        """
        Calculate the classification scores according to a set of predicted values (`pred`) and actual values (`actual`).
        If `labels.size > 2`, a `MultinomialScores` will be returned. Otherwise it will return `BinomialScores`.

        Parameters
        ----------
        pred : np.ndarray
            Prediction results based on test set feature columns
        pred_proba : np.ndarray
            Nested list of probability scores.

            Each sub-list represents the probability of each class being the
            proper prediction for a particular record. Used for the log loss metric
            e.g., `[[ 0.1, 0.8, 0.1] [ 0.0, 0.79 , 0.21] .... ]`,
            meaning that, for record X[0] in the test set, the proper prediction
            being class_0 is 10%, class_1 is 80%, and class_2 is 10%.
        actual : np.ndarray
            The test set target values from the original dataframe
        labels : np.ndarray
            The labels that appear in target column.

        Returns
        -------
        An instance of either `MultinomialScores` or `BinomialScores`.
        """

        pred, actual = sanitize_pred_and_actual(pred, actual)

        if labels.size < 2:
            raise ValueError(f"Invalid labels, has to be more than 1, got {labels}")

        def mean_per_class_error(pred: List[float], actual: List[float]):
            from collections import Counter, defaultdict

            element_count = dict(Counter(actual))
            misclassification_count: DefaultDict = defaultdict(int)
            for y, yhat in zip(pred, actual):
                if y != yhat:
                    misclassification_count[yhat] += 1
            return sum(
                [
                    misclassification_count[i] / float(element_count[i])
                    for i in element_count
                ]
            ) / len(element_count)

        def lift_top_group(pred: List[float], actual: List[int]) -> float:
            def mean(arr):
                return sum(arr) / float(len(arr))

            top_group = sorted(zip(pred, actual), key=lambda x: x[0])
            one_percent_len = int(math.ceil(len(top_group) / 100.0))
            return mean([x[1] for x in top_group[:one_percent_len]]) / float(
                mean(actual)
            )

        le = preprocessing.LabelEncoder()
        labels = le.fit_transform(labels)
        actual = le.transform(actual)
        pred = le.transform(pred)
        log_loss_ = log_loss(actual, pred_proba, labels=labels)
        # pred_scores = [pred_proba[i[0]][x] for i, x in np.ndenumerate(pred)]
        misclassification = 1 - accuracy_score(actual, pred)
        mpce = mean_per_class_error(list(pred), list(actual))
        if len(labels) > 2:  # n-ary classification problem
            return MultinomialScores(
                logloss=log_loss_,
                mean_per_class_error=mpce,
                misclassification=misclassification,
            )
        else:
            pred_scores = [pred_proba[i][1] for i in range(len(pred_proba))]
            fpr, tpr, _ = roc_curve(actual, pred_scores, pos_label=None)
            auc_ = auc(fpr, tpr)
            lift = lift_top_group(pred_scores, list(actual))
            return BinomialScores(
                logloss=log_loss_,
                mean_per_class_error=mpce,
                misclassification=misclassification,
                auc=auc_,
                lift_top_group=lift,
            )


class BinomialScores(ClassificationScores):
    """
    The binomial scores set.

    Metrics include:
        - logloss (`logloss`)
        - misclassification (`misclassification`)
        - mean per class error (`mean_per_class_error`)
        - area under curve (`auc`)
        - lift top group (`lift_top_group`)
    """

    auc: float
    lift_top_group: float

    @validator("auc")
    def check_auc(cls, v):
        if v != np.nan and v < 0:
            raise ValueError(f"{v} should be greater than or equal to 0")
        return v


class MultinomialScores(ClassificationScores):
    """
    The multinomial scores set.

    Metrics include:
        - logloss (`logloss`)
        - misclassification (`misclassification`)
        - mean per class error (`mean_per_class_error`)
    """


Scores = Union[RegressionScores, BinomialScores, MultinomialScores]


class CVScores(BaseModel):
    """
    The CV scores struct.

    Usage
    -----
    ```python
    split_scores: List[Scores] = []
    for _ in nfolds:
        # dates, actual = extract_from_data(df)
        # pred = model.predict(dates)
        split_scores.append(RegressionScores.get_scores(pred, actual))
    # To get the CV scores that can be sent from FastAPI, do this
    CVScores(split_scores=split_scores).to_report()
    ```
    """

    split_scores: List[Scores]

    @validator("split_scores")
    def check_folds(cls, v):
        if not v:
            raise ValueError("No score")
        types = {type(f).__name__ for f in v}
        if len(types) > 1:
            raise ValueError(
                f"All split scores must have consistent type. Found {types}"
            )
        return v

    def to_report(self) -> Dict[str, List[float]]:
        """
        Turns this CV scores object into a `dict` from `str` (the metric names) to a list of `floats` (the actual scores).

        """
        return {
            field: [score.dict()[field] for score in self.split_scores]
            for field in self.split_scores[0].__fields__
        }



import math
from io import BytesIO
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd

from exodusutils.constants import DATETIME_COMPONENTS_PREFIX
from exodusutils.enums import DataType, TimeUnit
from exodusutils.exceptions import ExodusBadRequest
from exodusutils.schemas import Column


def cast_columns(
    df: pd.DataFrame, columns: List[str], target_type: DataType
) -> pd.DataFrame:
    """
    Casts the specified columns to the target `DataType`.
    """
    for col in columns:
        if target_type == DataType.double:
            df[col] = pd.to_numeric(df[col], errors="coerce")
        elif target_type == DataType.timestamp:
            df[col] = pd.to_datetime(pd.Series(df[col]), errors="coerce")
        else:
            df[col] = df[col].fillna("").astype(str).replace("", np.nan)
    return df


def cast_df_types(df: pd.DataFrame, feature_types: List[Column]) -> pd.DataFrame:
    feature_types = [f for f in feature_types if f.name in df.columns]
    numeric_columns = [f.name for f in feature_types if f.data_type == DataType.double]
    datetime_columns = [
        f.name for f in feature_types if f.data_type == DataType.timestamp
    ]
    categorical_columns = [
        f.name
        for f in feature_types
        if f.name not in numeric_columns + datetime_columns
    ]
    return (
        df.pipe(cast_columns, numeric_columns, DataType.double)
        .pipe(cast_columns, datetime_columns, DataType.timestamp)
        .pipe(cast_columns, categorical_columns, DataType.string)
    )


def get_df(data: bytes, feature_types: List[Column]) -> pd.DataFrame:
    return pd.DataFrame(
        pd.read_csv(BytesIO(data), usecols=[f.name for f in feature_types])
    ).pipe(cast_df_types, feature_types)


def postfix_datetime_cols_with_time_components(col: str) -> List[str]:
    """
    Returns a list of str, representing the new time component names

    Parameters
    ----------
    col : str
        a datetime column

    Returns
    -------
    List[str]
        A list of new column names
    """
    return [f"{col}_{t}" for t in DATETIME_COMPONENTS_PREFIX]


def get_time_components(
    df: pd.DataFrame, time_unit: TimeUnit, col: str
) -> List[Optional[pd.Series]]:
    """
    Returns a list of `pd.Series`s, representing the time components for the given column name. \
A time component is a series of `np.int64`.

    Parameters
    ----------
    df : pd.DataFrame
        Original dataframe that features are to be derived from
    time_unit : TimeUnit
        The `smallest` time unit that the new features should include
    col : str
        One of the datetime columns of df

    Returns
    -------
    List of time components.
    """
    # Time components
    dt = df[col].dt
    components = [dt.year]

    # quarter, month and week component
    if time_unit not in {TimeUnit.year}:
        components += [dt.quarter, dt.month, dt.week]
    else:
        components += [None, None, None]

    # weekday and date component
    if time_unit not in {TimeUnit.year, TimeUnit.month}:
        components += [dt.weekday]
    else:
        components += [None]

    # hour component
    if time_unit not in {TimeUnit.year, TimeUnit.month, TimeUnit.day}:
        components += [dt.hour.replace(0, 24), dt.minute, dt.second]
    else:
        components += [None, None, None]

    return components


def get_columns(df: pd.DataFrame, dtype) -> List[str]:
    return df.select_dtypes(dtype).columns.tolist()


def merge_training_holdout(
    training: pd.DataFrame, holdout: pd.DataFrame
) -> pd.DataFrame:
    holdout["holdout"] = 1
    training["holdout"] = 0
    return pd.concat([training, holdout])


def split_to_training_and_holdout(
    df: pd.DataFrame,
) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
    if "holdout" not in df.columns:
        return df, None
    else:

        def drop_holdout(df: pd.DataFrame, val: int) -> pd.DataFrame:
            return df.loc[df["holdout"] == val, :].drop(columns="holdout")

        return drop_holdout(df, 0), drop_holdout(df, 1)


def get_numeric_features(df: pd.DataFrame, target: str) -> List[str]:
    return [
        f
        for f in get_columns(df, np.float64) + get_columns(df, np.int64)
        if f != target
    ]


def train_validation_split(
    df: pd.DataFrame, validation_percentage: float
) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
    if validation_percentage == 0.0:
        return df, None
    else:
        indexed_validation = df.sample(frac=validation_percentage)
        return df.loc[
            ~df.index.isin(indexed_validation.index)
        ], indexed_validation.reset_index(drop=True)


def sanitize_pred_and_actual(
    pred: np.ndarray, actual: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    if pred.size != actual.size:
        raise ValueError(f"pred.size = {pred.size} != actual.size = {actual.size}")

    if np.issubdtype(actual.dtype, np.number):
        if not np.issubdtype(pred.dtype, np.number):
            raise ValueError(f"actual is number while pred is not")
        pred, actual = [
            np.array(t)
            for t in zip(*[(p, a) for p, a in zip(pred, actual) if not np.isnan(a)])
        ]

    if pred.size == 0:
        raise ValueError("empty fold score")

    return pred, actual


def validate_columns(header: List[str], features: List[Column]) -> None:
    """
    Check if there are any required features in `features` that are missing from the dataframe

    Parameters
    ----------
        header : List[str]
            columns that are present in a dataframe
        features : List[Column]
            `features` is a list of `Column`s that is/was required for training

    Returns
    -------
        None

    Raises
    ------
        HTTPException
            the client is notified that the information supplied is invalid
    """
    for f in features:
        if (
            f.name not in header
        ):  # This means that a required field is missing from the dataframe
            raise ExodusBadRequest(f"column {f.name} is missing from the data")


def get_column_mode(col) -> float:
    """
    Returns the most frequent value of a numeric column, or `0.0` if it is impossible to calculate.

    This is used to create labels for the previously unseen values in a label encoded categorical column.
    """
    mode = col.mode().values
    if len(mode) > 0:
        return (
            0.0 if math.isnan(mode[0]) else mode[0]
        )  # Just take the first one if there are ties
    else:
        return 0.0


def append_fold_column(df: pd.DataFrame, name: str, folds: int) -> pd.DataFrame:
    """Appends the fold column to the incoming dataframe. Will return a new dataframe, and leave the old one unchanged.

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe.
    name : str
        Name of the fold column.
    folds : int
        Number of folds we want to cut.

    Returns
    -------
    pd.DataFrame
        The resulting dataframe.

    """
    res = df.copy()
    res[name] = np.array((df.index % folds).astype(int))
    return res


def select_and_drop_fold_column(
    df: pd.DataFrame, name: str, fold: int, test: bool
) -> pd.DataFrame:
    """
    Selects the rows that are in fold range, and then drops the fold column.

    If `test` is `True`, selects the rows with `name` equal to `fold`. Otherwise chooses the ones that are not equal.

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe.
    name : str
        Name of fold column.
    fold : int
        The fold to select.
    test : bool
        Whether we are selecting test frame or not.

    Returns
    -------
    pd.DataFrame
        The selected dataframe.

    """
    return (
        df.loc[(df[name] == fold) if test else (df[name] != fold)]
        .reset_index(drop=True)
        .drop(columns=[name])
    )



from typing import List, Optional

import numpy as np
import pandas as pd

from exodusutils.exceptions.exceptions import ExodusMethodNotAllowed
from exodusutils.internal.process_unit import ProcessUnit
from exodusutils.schemas import Column


class ConstantColumnsRemover(ProcessUnit):
    """
    Locates all constant columns in the dataframe during `fit`, then during `transform` remove the constant \
columns found in `fit` from the given dataframe. Will not remove the target column.
    """

    def __init__(self, feature_types: List[Column], target_column_name: str) -> None:
        self.feature_types = feature_types
        self.target_column_name = target_column_name
        self.constant_columns: Optional[List[str]] = None

    def fitted(self) -> bool:
        return self.constant_columns is not None

    def fit(self, df: pd.DataFrame) -> None:
        self.constant_columns = [
            column.name
            for column in self.feature_types
            if column.name != self.target_column_name
            and (
                (df[column.name].isnull().all())  # All NaNs
                or (
                    df[column.name].notnull().all() and df[column.name].nunique() == 1
                )  # Only one kind of value without any NaN
            )
        ]

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        if not self.fitted():
            raise ExodusMethodNotAllowed("Has to fit before transform")
        return pd.DataFrame(df.drop(columns=self.constant_columns))



from typing import Dict

import numpy as np
import pandas as pd

from exodusutils.constants import PREDICTION_COLNAME


def append_predictions(
    df: pd.DataFrame,
    predictions: np.ndarray,
    class_predict_probas: Dict[str, np.ndarray],
) -> pd.DataFrame:
    """
    Append predictions to the dataframe.

    If this is a classification experiment, then supply this method with `class_predict_probas`, a
    mapping from class names to class probabilities.

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe
    predictions : np.ndarray
        predictions
    class_predict_probas : Dict[str, np.ndarray]
        class predict probailities

    Returns
    -------
    pd.DataFrame

    """
    df[PREDICTION_COLNAME] = predictions
    for class_name, class_proba in class_predict_probas.items():
        df[class_name] = class_proba
    return df



from exodusutils.feature_engineering.label_encoding import LabelEncoding
from exodusutils.feature_engineering.one_hot_encoding import OneHotEncoding
from exodusutils.feature_engineering.time_component_encoding import (
    TimeComponentEncoding,
)



from typing import List, Optional

import numpy as np
import pandas as pd

from exodusutils import internal


def to_numeric_features_df(df: pd.DataFrame, target: str) -> pd.DataFrame:
    """
    Extracts the numeric features from the dataframe, and returns them as a new dataframe.

    The numeric features include:
        - Columns declared with `data_type == DataType.double` in `feature_types`
        - The encoded columns (i.e. the dummy columns in one-hot encoding, or the
                columns that have been label encoded)

    This is the `X` dataframe used in TPOT.

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe.
    target : str
        Target column name.

    Returns
    -------
    The numeric features as a dataframe.
    """
    return pd.DataFrame(df[internal.get_numeric_features(df, target)])


def remove_invalid_values(df: pd.DataFrame) -> pd.DataFrame:
    """
    Removes the invalid values from the dataframe, and resets the index.

    The invalid values are:
    - `float("nan")` or `np.nan` in numeric columns
    - `""` in other types of columns
    """
    return pd.DataFrame(df.replace("", np.nan).dropna()).reset_index(drop=True)


def remove_datetime_columns(df: pd.DataFrame) -> pd.DataFrame:
    """
    Removes all columns of dtype `datetime64`.
    """
    return pd.DataFrame(df.drop(internal.get_columns(df, np.datetime64), axis=1))


def fill_nan_with_mode(df: pd.DataFrame, columns: List[str]) -> pd.DataFrame:
    """
    Fills the `NaN` cells in each numeric column with the most frequent value (aka `mode`) of that column.

    This is used to create labels for the previously unseen values in a label encoded categorical column.

    Parameters
    ----------
    df : DataFrame
        The dataframe to fill.
    columns : List[str]
        The names of the label encoded columns to fill.

    Returns
    -------
    DataFrame
        The dataframe with no `NaN` cell.
    """
    value = {col: internal.get_column_mode(df[col]) for col in columns}
    return df.fillna(value=value)


def drop_blank(df: pd.DataFrame, target_column_name: Optional[str]) -> pd.DataFrame:
    """
    Drops the blank rows from the dataframe. If target column name is given, then all rows with `np.nan` or `\"\"` \
in the target column will be discarded. Otherwise only rows with all `np.nan` or `\"\"` will be discarded.

    Parameters
    ----------
    df : DataFrame
        The dataframe to manipulate on.
    target_column_name : Optional[str]
        The target column name. Optional.

    Returns
    -------
    DataFrame
        The dataframe without blank rows.
    """
    if target_column_name is not None:
        dropped = df.dropna(subset=[target_column_name])
    else:
        dropped = pd.DataFrame(df.dropna(how="all"))
    return dropped.reset_index(drop=True)



from typing import List, Optional

import pandas as pd
from pydantic import BaseModel

from exodusutils import internal
from exodusutils.exceptions.exceptions import ExodusMethodNotAllowed


class SplitFrames(BaseModel):
    """
    A collection of dataframes. The validation frame and test frame are optional.
    """

    train: pd.DataFrame
    validation: Optional[pd.DataFrame]
    test: Optional[pd.DataFrame]

    class Config:
        arbitrary_types_allowed = True


class CVFrames(BaseModel):
    """
    A list of `TrainFrames`. User should not initialize this class directly, instead they should \
use the classmethod `iid`.
    """

    splits: List[SplitFrames]

    @classmethod
    def iid(
        cls,
        df: pd.DataFrame,
        nfolds: int,
        validation_df: Optional[pd.DataFrame] = None,
        fold_assignment_column_name: Optional[str] = None,
    ):
        """
        Create CV frames for an IID experiment.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to split to CV frames.
        nfolds : int
            Number of folds.
        validation_df : Optional[pd.DataFrame]
            The dataframe for validation. Optional.
        fold_assignment_column_name : Optional[str]
            If defined, fold will be cut according to it.

        Returns
        -------
        CVFrames
        """
        splits = []

        if fold_assignment_column_name is None:
            name: str = "FOLD_COLUMN"
            df = internal.append_fold_column(df, name, nfolds)
        else:
            if fold_assignment_column_name not in df.columns:
                raise ExodusMethodNotAllowed(
                    f"Fold column {fold_assignment_column_name} not found in dataframe"
                )
            name = fold_assignment_column_name

        for fold in range(nfolds):
            train = internal.select_and_drop_fold_column(df, name, fold, test=False)
            test = internal.select_and_drop_fold_column(df, name, fold, test=True)
            splits.append(SplitFrames(train=train, validation=validation_df, test=test))
        return cls(splits=splits)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['exodusutils',
 'exodusutils.exceptions',
 'exodusutils.feature_engineering',
 'exodusutils.internal',
 'exodusutils.migration',
 'exodusutils.predict',
 'exodusutils.preprocessing',
 'exodusutils.schemas']

package_data = \
{'': ['*']}

install_requires = \
['colorama>=0.4.4,<0.5.0',
 'fastapi>=0.74.1,<0.75.0',
 'numpy==1.21',
 'pandas>=1.4.0,<2.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'pymongo[encryption]>=4.0.1,<5.0.0',
 'python-Levenshtein>=0.12.2,<0.13.0',
 'scikit-learn>=1.0.2,<2.0.0']

setup_kwargs = {
    'name': 'exodusutils',
    'version': '0.1.29',
    'description': 'Utility functions and helper classes for Exodus project',
    'long_description': "# Exodus common utilities\n\nThis is the library defining the schemas for exodus utilities.\n\n## Structure\n\n### `exodusutils`\n\nContains helpful utility functions.\n\n### `schemas`\n\nIn the `schemas` folder you can find the following:\n- Schema definitions for the model algorithm's incoming requests\n- Schema definitions for the model algorithm's responses\n- Definitions for `RegressionScores` and `ClassificationScores`\n- Definitions for types such as `Attribute` and `Column`\n\n### `predict`\n\nThe `predict` folder contains helper functions for prediction.\n\n### `enums.py`\n\nContains enums used by Exodus. Current contains the following:\n- `TimeUnit`, with helper methods to convert timestamps to different formats\n- `DataType`, with helper methods to convert from `Pandas` types\n\n### `feature_engineering.py`\n\nContains commonly used feature engineering methods. Currently includes:\n- One-hot encoding\n- Label encoding\n- Time component encoding\nIt is recommended to use at least 1 method in this file during training.\n\n### `frame_manipulation.py`\n\nContains multiple frame manipulation methods. Used during prediction, should pick the method that corresponds to the one used during training.\n\n### `frames.py`\n\nContains definitions and helper functions for the following classes:\n- `SplitFrames`: A 3-tuple with a training dataframe, a testing dataframe, and a validation dataframe.\n- `CVFrames`: A list of `SplitFrames`. Should not be instantiated manually, user should use `CVFrames.iid` helper classmethod.\n",
    'author': 'Tsung-Ju Lii',
    'author_email': 'andylii@mobagel.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.0,<4.0.0',
}


setup(**setup_kwargs)



import abc
from typing import Optional, Tuple

import pandas as pd


class ProcessUnit(abc.ABC):
    def fitted(self) -> bool:
        """Whether this process unit has been fitted or not."""
        raise NotImplementedError

    def fit(self, df: pd.DataFrame) -> None:
        """Fits this processing unit based on the given dataframe.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to fit this processing unit.

        Returns
        -------
        None

        """
        raise NotImplementedError

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        """Transforms the given dataframe based on the results stored in this processing unit after calling `fit`.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to transform.

        Returns
        -------
        pd.DataFrame
            The transformed dataframe.

        """
        raise NotImplementedError

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        """Fits this processing unit based on the dataframe, and then transform the dataframe based on the fit results.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to fit this processing unit and transformed.

        Returns
        -------
        pd.DataFrame
            The transformed dataframe.

        """
        self.fit(df)
        return self.transform(df)

    def maybe_transform(
        self, maybeDf: Optional[pd.DataFrame]
    ) -> Optional[pd.DataFrame]:
        """Transforms the input dataframe if it is not a `None`."""
        return None if maybeDf is None else self.transform(maybeDf)

    def run(
        self,
        training_df: pd.DataFrame,
        validation_df: Optional[pd.DataFrame],
        holdout_df: Optional[pd.DataFrame],
    ) -> Tuple[pd.DataFrame, Optional[pd.DataFrame], Optional[pd.DataFrame]]:
        """
        Will do the following:
        - `fit_transform` on the training dataframe.
        - `transform` on the validation dataframe if it is not a `None`.
        - `transform` on the holdout dataframe if it is not a `None`.
        - Return the dataframes.
        """
        training_df = self.fit_transform(training_df)
        validation_df = self.maybe_transform(validation_df)
        holdout_df = self.maybe_transform(holdout_df)
        return training_df, validation_df, holdout_df
